package ru.dbotthepony.mc.otm.player.android.feature

import net.minecraft.client.multiplayer.ClientLevel
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.codec.StreamCodec
import net.minecraft.network.protocol.common.custom.CustomPacketPayload
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.item.ItemEntity
import net.neoforged.neoforge.network.PacketDistributor
import net.neoforged.neoforge.network.handling.IPayloadContext
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.config.PlayerConfig
import ru.dbotthepony.mc.otm.player.android.AndroidSwitchableFeature
import ru.dbotthepony.mc.otm.player.MatteryPlayer
import ru.dbotthepony.mc.otm.capability.energy.extractEnergyExact
import ru.dbotthepony.mc.otm.client.render.MGUIGraphics
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.ResearchIcons
import ru.dbotthepony.mc.otm.core.ResourceLocation
import ru.dbotthepony.mc.otm.core.math.Vector
import ru.dbotthepony.mc.otm.core.getEntitiesInEllipsoid
import ru.dbotthepony.mc.otm.core.math.minus
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.position
import ru.dbotthepony.mc.otm.core.math.times
import ru.dbotthepony.mc.otm.registry.game.AndroidFeatures
import java.util.UUID
import java.util.WeakHashMap
import java.util.function.Predicate

private data class SharedItemEntityData(val owner: UUID? = null, val age: Int = 0, val lifespan: Int = 0, val hasPickupDelay: Boolean = true) {
	companion object {
		val EMPTY = SharedItemEntityData()
	}
}

private val datatable = WeakHashMap<ItemEntity, SharedItemEntityData>()

class ItemEntityDataPacket(val itemUUID: Int, val owner: UUID? = null, val age: Int = 0, val lifespan: Int = 0, val hasPickupDelay: Boolean = true) : CustomPacketPayload {
	fun write(buff: FriendlyByteBuf) {
		buff.writeVarInt(itemUUID)
		buff.writeBoolean(owner != null)
		if (owner != null) buff.writeUUID(owner)
		buff.writeVarInt(age)
		buff.writeVarInt(lifespan)
		buff.writeBoolean(hasPickupDelay)
	}

	fun play(context: IPayloadContext) {
		val level = minecraft.player?.level() as ClientLevel? ?: return
		val entity = level.getEntity(itemUUID) as ItemEntity? ?: return
		datatable[entity] = SharedItemEntityData(owner, age, lifespan, hasPickupDelay)
	}

	override fun type(): CustomPacketPayload.Type<out CustomPacketPayload> {
		return TYPE
	}

	companion object {
		val TYPE = CustomPacketPayload.Type<ItemEntityDataPacket>(ResourceLocation(OverdriveThatMatters.MOD_ID, "item_entity_data"))
		val CODEC: StreamCodec<FriendlyByteBuf, ItemEntityDataPacket> = StreamCodec.ofMember(
			ItemEntityDataPacket::write,
			Companion::read
		)

		fun read(buff: FriendlyByteBuf): ItemEntityDataPacket {
			return ItemEntityDataPacket(buff.readVarInt(), if (buff.readBoolean()) buff.readUUID() else null, buff.readVarInt(), buff.readVarInt(), buff.readBoolean())
		}
	}
}

class ItemMagnetFeature(capability: MatteryPlayer) : AndroidSwitchableFeature(AndroidFeatures.ITEM_MAGNET, capability) {
	private data class ItemPos(var position: Vector, var ticksSinceActivity: Int)
	private val rememberPositions = WeakHashMap<ItemEntity, ItemPos>()

	private val serverPredicate = Predicate<Entity> { it is ItemEntity && !it.hasPickUpDelay() && (it.owner == null || it.owner != ply || it.lifespan - it.age <= 200) }
	private val clientPredicate = Predicate<Entity> { it is ItemEntity && (datatable[it] ?: SharedItemEntityData.EMPTY).let { !it.hasPickupDelay && (it.owner == null || it.owner != ply.uuid || it.lifespan - it.age <= 200) } }

	private fun doTick(server: Boolean) {
		if (ply.isSpectator || server && !android.androidEnergy.extractEnergyExact(PlayerConfig.Magnet.POWER_DRAW, true)) {
			return
		}

		val entities = ply.level().getEntitiesInEllipsoid(
			ply.position,
			Vector(PlayerConfig.Magnet.RADIUS_HORIZONTAL, PlayerConfig.Magnet.RADIUS_VERTICAL, PlayerConfig.Magnet.RADIUS_HORIZONTAL),
			if (server) Predicate<Entity> { it is ItemEntity } else clientPredicate
		)

		for ((ent, distance) in entities) {
			ent as ItemEntity

			if (server) {
				PacketDistributor.sendToPlayer(ply as ServerPlayer,  ItemEntityDataPacket(ent.id, ent.owner?.uuid, ent.age, ent.lifespan, ent.hasPickUpDelay()))

				if (!serverPredicate.test(ent)) {
					continue
				}

				val data = rememberPositions.computeIfAbsent(ent) { ItemPos(it.position, 0) }

				if (data.position.distanceToSqr(ent.position) < 1.0) {
					data.ticksSinceActivity++
				} else {
					if (!android.androidEnergy.extractEnergyExact(PlayerConfig.Magnet.POWER_DRAW, false)) {
						return
					}

					data.position = ent.position
					data.ticksSinceActivity = 0
				}
			}

			ent.deltaMovement += (ply.position - ent.position).normalize() * ((1.0 - distance) * 0.2)
		}
	}

	override fun tickClient() {
		super.tickClient()

		if (!ply.isSpectator && isActive && android.androidEnergy.extractEnergyExact(PlayerConfig.Magnet.POWER_DRAW, true)) {
			doTick(false)
		}
	}

	override fun tickServer() {
		super.tickServer()

		if (isActive) {
			doTick(true)
		}
	}

	override fun renderIcon(graphics: MGUIGraphics, x: Float, y: Float, width: Float, height: Float, color: RGBAColor) {
		ResearchIcons.ICON_ITEM_MAGNET.render(graphics, x, y, width, height, color = color)
	}
}
